package simFil;
/* FileOps... like it sounds */

/*  <SimFil - simulation of cytoskeletal filament flexibility by Pairwise Agent Interaction with Rational Superposition (PAIRS)>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import infoCCD.Info;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.Math;
import java.awt.*;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.JFileChooser;
import javax.swing.JFrame;

public class FileOps_ {
	// type definitions
	static final int NOTHING = 999;
	static final int TIME = -1;
	static final int PARM_CONCENTRATION = -2;
	static final int NUM_FILS = -3;
	static final int NUM_MONS = -4;
	static final int NUM_FILSEGS = -5;
	static final int NUM_FILLINKS = -6;
	static final int NUM_PLASMIDS = -7;
	static final int NUM_THINGS = 0;
	static final int FILSEGMENT = 1;
	static final int PLASMID = 2;
	static final int BUG = 3;
	static final int FILLINK = 4;
	
	static String sepString = ";";
	static String divStr = ":";
	static String endStr = ";";
	static String idDivStr = ":";
	static String commentStr = "//";
	// file, file writer, and print writer for program errors to file
	static File errorsOutFile;
	static FileOutputStream errorsOutFOS;
	static PrintStream errorsOutPS;	
	// file, file writer, and print writer for program messages to file
	static File messagesOutFile;
	static FileWriter messagesOutFW;
	static PrintWriter messagesOutPW;	
	// file, file writer, and print writer for information file
	static File outFile;
	static FileWriter outFW;
	static PrintWriter outPW;
	// file, file writer, and print writer for persistence length file
	static File lpFile;
	static FileWriter lpFW;
	static PrintWriter lpPW;
//	 file, file writer, and print writer for apparent diffusivity file
	static File aDFile;
	static FileWriter aDFW;
	static PrintWriter aDPW;
	// files and writers for filament lengths file
	static File filLengthFile;
	static FileWriter filLengthFW;
	static PrintWriter filLengthPW;
	// files and writers for filament length histogram file
	static File filLengthHistoFile;
	static FileWriter filLengthHistoFW;
	static PrintWriter filLengthHistoPW;
	// files and writers for filament length histogram file
	static File relaxFile;
	static FileWriter relaxFileFW;
	static PrintWriter relaxFilePW;
	// flags for file setup
	static boolean outFileMade = false;
	static boolean lpFileMade = false;
	static boolean aDFileMade = false;
	static boolean filLengthFileMade = false;
	static boolean filLengthHistoFileMade = false;
	static boolean relaxFileMade = false;
	
	// file id formating
	static DecimalFormat fileIdFormat = new DecimalFormat ("#000000000.#;#000000000.#");


	public FileOps_ () {
	}
	

	// A class to filter out .java files
	class SourceFileFilter implements FilenameFilter {

    /**
    * Select only .java files.
    *
    * @param dir the directory in which the file was found.
    *
    * @param name the name of the file
    *
    * @return true if and only if the name should be
    * included in the file list; false otherwise.
    */
    public boolean accept ( File dir, String name )
      {
      if ( new File( dir, name ).isDirectory() )
         {
         return false;
         }
      name = name.toLowerCase();
      return name.endsWith( ".java" );
      }
    }
	
	//	 A class to filter out .param files
	class ParamFileFilter implements FilenameFilter {

    /**
    * Select only .param files.
    *
    * @param dir the directory in which the file was found.
    *
    * @param name the name of the file
    *
    * @return true if and only if the name should be
    * included in the file list; false otherwise.
    */
    public boolean accept ( File dir, String name )
      {
      if ( new File( dir, name ).isDirectory() )
         {
         return false;
         }
      name = name.toLowerCase();
      return name.endsWith( ".param" );
      }
    }
	
	// A class to filter out .qk files
	static class QKFileFilter implements FilenameFilter {

    /**
    * Select only .qk files.
    *
    * @param dir the directory in which the file was found.
    *
    * @param name the name of the file
    *
    * @return true if and only if the name should be
    * included in the file list; false otherwise.
    */
    public boolean accept ( File dir, String name )
      {
      if ( new File( dir, name ).isDirectory() )
         {
         return false;
         }
      name = name.toLowerCase();
      return name.endsWith( ".qk" );
      }
    }
    
    public void copyFile(File in, File out) throws Exception {
	    FileInputStream fis  = new FileInputStream(in);
	    FileOutputStream fos = new FileOutputStream(out);
	    byte[] buf = new byte[1024];
	    int i = 0;
	    while((i=fis.read(buf))!=-1) {
	      fos.write(buf, 0, i);
	    }
	    fis.close();
	    fos.close();
	}
    
    public static String [] getQKFileList (File dir) {
    		return dir.list(new QKFileFilter());
    }
    
	public static String getDirectoryToSave (String dialogTitle, Component parent) {
		File file = null;
		String fileName = null;
		JFileChooser saveTo = new JFileChooser();
		saveTo.setDialogTitle(dialogTitle);
		saveTo.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		int fileQuery = saveTo.showSaveDialog(parent);
		if (fileQuery == JFileChooser.APPROVE_OPTION) {
            file = saveTo.getSelectedFile();
            if (!file.exists()) { file.mkdir(); }
            fileName = file.getAbsolutePath();
           // talkln (dialogTitle + fileName);
        } else {
            talkln ("Save cancelled by user");
            return null;
        }
        return fileName;
	}
	
	public static String getFileToSave (String dialogTitle, JFrame parent) {
		File file = null;
		String fileName = null;
		JFileChooser saveTo = new JFileChooser();
		saveTo.setDialogTitle(dialogTitle);
		int fileQuery = saveTo.showSaveDialog(parent);
		if (fileQuery == JFileChooser.APPROVE_OPTION) {
            file = saveTo.getSelectedFile();
            fileName = file.getAbsolutePath();
            //talkln (dialogTitle + fileName);
        } else {
            talkln ("Save cancelled by user");
            return null;
        }
        return fileName;
	}
	
	public static String getDirectoryToLoad (String dialogTitle, JFrame parent) {
		File file = null;
		String fileName = null;
		JFileChooser loadFrom = new JFileChooser();
		loadFrom.setDialogTitle(dialogTitle);
		loadFrom.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		int fileQuery = loadFrom.showOpenDialog(parent);
		if (fileQuery == JFileChooser.APPROVE_OPTION) {
            file = loadFrom.getSelectedFile();
            fileName = file.getAbsolutePath();
            //talkln (dialogTitle + fileName);
        } else {
            talkln ("Load cancelled by user");
            return null;
        }
        return fileName;
	}
	
	public static String getFileToLoad (String dialogTitle, JFrame parent) {
		File file = null;
		String fileName = null;
		JFileChooser loadFrom = new JFileChooser();
		loadFrom.setDialogTitle(dialogTitle);
		loadFrom.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
		int fileQuery = loadFrom.showOpenDialog(parent);
		if (fileQuery == JFileChooser.APPROVE_OPTION) {
            file = loadFrom.getSelectedFile();
            fileName = file.getAbsolutePath();
            //talkln (dialogTitle + fileName);
        } else {
            talkln ("Load cancelled by user");
            return null;
        }
        return fileName;
	}
	
	public static String makeNameFromIntStep (String base, int integrationStep) {
		return (base + String.valueOf(fileIdFormat.format(integrationStep)));
	}
	
	public void mkErrorFile () {
		try {
			errorsOutFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Errors");
			errorsOutFOS = new FileOutputStream(errorsOutFile);
			errorsOutPS = new PrintStream(errorsOutFOS);
			System.setErr(errorsOutPS);
			//Date timeNow = new Date();
			//errorsOutPW.println("This is the errors file for " + Env.outFileName + " run on " + timeNow.toString());
		} catch (IOException ioe) { talkln ("An error creating errorsOut stuff"); }
	}
		
	public void mkMessageFile () {
		try {
			messagesOutFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Messages");
			messagesOutFW = new FileWriter(messagesOutFile);
			messagesOutPW = new PrintWriter(messagesOutFW, true);
			Date timeNow = new Date();
			messagesOutPW.println("This is the message file for " + Env.outFileName + " run on " + timeNow.toString());
		} catch (IOException ioe) { talkln ("An error creating messagesOut stuff"); }
	}
	
	public void mkEnvironmentFile () {
		File envFile = new File ("Env.java");
		if ( ! envFile.exists() || ! envFile.canRead() ) {
			talkln ("Can't find 'Env.java'... you'd better store those values manually");
		} else {
			try {
				FileReader fileR = new FileReader(envFile);
				BufferedReader in = new BufferedReader(fileR);
				File envOut = new File(Env.logFolderPath + File.separator + Env.outFileName + "Environment");
				FileWriter envOutFW = new FileWriter(envOut);
				PrintWriter envOutPW = new PrintWriter(envOutFW, true);
				Date timeNow = new Date();
				envOutPW.println("This is the environment file for " + Env.outFileName + " run on " + timeNow.toString());
				String nextLine = in.readLine();
				while (nextLine != null) {
					envOutPW.println(nextLine);
					nextLine = in.readLine();
				}
			} catch (IOException ioe) { talkln ("An error trying to save environment file"); }
		}
	}
	
	public void mkSourceFolder () {
		try {
			File curDir = new File (".");
			System.out.println ("Directory for source files is: " + curDir.getCanonicalPath());
			String [] srcFiles = curDir.list(new SourceFileFilter());
			for (int i=0;i<srcFiles.length;i++) {
				File curSrcFile = new File (srcFiles[i]);
				File curOutSrcFile = new File (Env.srcFilePath + File.separator + curSrcFile.getName());
				copyFile(curSrcFile,curOutSrcFile);
			}
		} catch (Exception e) { talkln ("An error trying to copy source files"); }
	}
	
	public static void mkOutFile () {
		try {
			outFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Out");
			outFW = new FileWriter(outFile);
			outPW = new PrintWriter(outFW,true);
			Date timeNow = new Date();
			outPW.println("This is the out file for " + Env.outFileName + " run on " + timeNow.toString());
			outPW.print("time;");
			
			outPW.println("");
			outFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating out file stuff"); }
	}
	
	public static void mkPersistenceLengthFile (String fileName) {
		try {
			lpFile = new File (fileName);
			lpFW = new FileWriter(lpFile);
			lpPW = new PrintWriter(lpFW,true);
			Date timeNow = new Date();
			lpPW.println("This is the Lp file for " + Env.outFileName + " run on " + timeNow.toString());
			lpPW.print("!time arcLength->;");
			
			double initArc = 0;//Env.stdSegLength.getIntValue()*Env.parMHalfMonoSize/2.0;
			double arcDelta = (Env.stdSegLength.getIntValue()+1)*Env.monomerDiameter.getValue()/2;
			for (int i=0;i<Env.persistenceSegNumber;i++) {
				double curArc = initArc + i*arcDelta;
				lpPW.print(curArc + ";");
			}
			lpPW.println("");
			lpFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating Lp file stuff"); }
	}
	
	public static void mkPersistenceLengthFile () {
		try {
			lpFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Lp");
			lpFW = new FileWriter(lpFile);
			lpPW = new PrintWriter(lpFW,true);
			Date timeNow = new Date();
			lpPW.println("This is the Lp file for " + Env.outFileName + " run on " + timeNow.toString());
			lpPW.print("!time arcLength->;");
			
			double initArc = 0;//Env.stdSegLength.getIntValue()*Env.parMHalfMonoSize/2.0;
			double arcDelta = (Env.stdSegLength.getIntValue()+1)*Env.monomerDiameter.getValue()/2;
			for (int i=0;i<Env.persistenceSegNumber;i++) {
				double curArc = initArc + i*arcDelta;
				lpPW.print(curArc + ";");
			}
			lpPW.println("");
			lpFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating Lp file stuff"); }
	}
	
	public static void mkDiffusivityFile () {
		try {
			aDFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Diffusivity");
			aDFW = new FileWriter(aDFile);
			aDPW = new PrintWriter(aDFW,true);
			Date timeNow = new Date();
			aDPW.println("This is the apparent diffusivity file for " + Env.outFileName + " run on " + timeNow.toString());
			aDPW.print("!time;");
			aDPW.print("xD;yD;zD;");
			aDPW.println("");
			aDFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating aD file stuff"); }
	}
	
	public static void mkFilLengthFile () {
		try {
			filLengthFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Filaments");
			filLengthFW = new FileWriter(filLengthFile);
			filLengthPW = new PrintWriter(filLengthFW,true);
			Date timeNow = new Date();
			filLengthPW.println("This is the filament lengths file for " + Env.outFileName + " run on " + timeNow.toString());
			filLengthPW.print("time;");
			filLengthPW.println("");
			filLengthFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating filament length file stuff"); }
	}
	
	public static void mkRelaxFile () {
		try {
			relaxFile = new File ("/Users/jonathan/Desktop/relaxFile");
			int j=1;
			while (relaxFile.isFile()) {
				System.out.println (relaxFile.getName() + " exists.... adding integers to end of name");
				relaxFile = new File("/Users/jonathan/Desktop/relaxFile." + String.valueOf(j));
				j++;
			}
			relaxFileFW = new FileWriter(relaxFile);
			relaxFilePW = new PrintWriter(relaxFileFW,true);
			Date timeNow = new Date();
			relaxFilePW.println("Relaxation file for length " + Env.filamentLength.getValue() + ", secondMode is " + Env.secondMode + " run on " + timeNow.toString());
			relaxFilePW.println("simTime;;");
			relaxFileMade = true;
		} catch (IOException ioe) { talkln ("An error creating relaxFile"); }
	}
	
	
	public static void writeToOutFile () {
		if (!outFileMade) { mkOutFile(); }
		printValue(Env.simulationTime,outPW);
		
		outPW.println("");
	}
	
	public static void writeToPersistenceLengthFile () {
		if (!lpFileMade) { mkPersistenceLengthFile(); }
		printValue(Env.simulationTime,lpPW);
		for (int i=0;i<Env.persistenceSegNumber;i++) {
			printValue(Env.persistenceTrack[i].runningAverageVal(),lpPW);
		}
		lpPW.println("");
	}
	
	public static void writeToDiffusivityFile () {
		if (!aDFileMade) { mkDiffusivityFile(); }
		printValue(Env.simulationTime,aDPW);
		// calculate D = variance/2*deltaT
		Pt3D curDPt = Env.transDiffTrack.varianceOfPt();
		curDPt.scale(1/(2*Env.ckPersistenceInterval*Env.deltaT.getValue()));
		printPt3D(curDPt,aDPW);
		aDPW.println("");
	}
	
	public static void writeToFilLengthFile () {
		if (!filLengthFileMade) { mkFilLengthFile(); }
		printValue(Env.simulationTime,filLengthPW);
		for (int i=0;i<FilSegment.filSegmentCt;i++) {
			filLengthPW.print(String.valueOf(FilSegment.theFilSegments[i].filID));
			filLengthPW.print(idDivStr);
			printValue(FilSegment.theFilSegments[i].length, filLengthPW);
		}
		filLengthPW.println("");
	}
	
	public static void writeToRelaxFile () {
		if (!relaxFileMade) { mkRelaxFile(); }
		printValue(Env.simulationTime,relaxFilePW);
		printValue(Env.currentDeflection,relaxFilePW);
		relaxFilePW.println();
	}
	
	private static void printDivider (PrintWriter pw) {
		pw.print(sepString);
	}
	
	private static void printValue (int val, PrintWriter pw) {
		pw.print(String.valueOf(val));
		printDivider(pw);
	}
	
	private static void printIntValue (double val, PrintWriter pw) {
		int intVal = (int) val;
		pw.print(String.valueOf(intVal));
		printDivider(pw);
	}
	
	private static void printValue (double val, PrintWriter pw) {
		pw.print(String.valueOf(val));
		printDivider(pw);
	}
	
	private static void printPt3D (Pt3D pt, PrintWriter pw) {
		pw.print(String.valueOf(pt.x));
		printDivider(pw);
		pw.print(String.valueOf(pt.y));
		printDivider(pw);
		pw.print(String.valueOf(pt.z));
		printDivider(pw);
	}
	
	public static void remoteParamConfigSave () {
		File configFile = new File (Env.logFolderPath + File.separator + Env.outFileName + "Params");
		writeParamConfig (configFile);
	}
	
	public static void writeParamConfig (File paramOutFile) {
		talk ("Saving parameter values to " + paramOutFile.getAbsolutePath() + "...");
		try {
			FileWriter paramOutFW = new FileWriter(paramOutFile);
			PrintWriter paramOutPW = new PrintWriter(paramOutFW, true);
			Date timeNow = new Date();
			paramOutPW.println(Info.appParamID + " This is a parameter configuration file created on " + timeNow.toString());
			Parameter_ curParam;
			String firstStr;
			for (int i=0;i<Parameter_.paramCt;i++) {
				curParam = Parameter_.theParams[i];
				firstStr = curParam.label + divStr + String.valueOf(curParam.isActive()) + divStr + String.valueOf(curParam.getValue()) + endStr;
				paramOutPW.print(firstStr);
				int spacesToAdd = 40-firstStr.length();
				for (int j=0;j<spacesToAdd;j++) { paramOutPW.print(" "); }
				paramOutPW.println(commentStr + curParam.parameterName);
			}
			talkln ("done.");
		} catch (IOException ioe) { talkln ("An error trying to write the parameter config file"); }
	}
	
	public static boolean loadParamLine (String curLine, String fromArg) {
		String curLabel,curValueStr,curActiveStr;
		boolean curActive = true;
		double curValue = 0;
		int div1Ind,div2Ind,endInd;
		// parse label and value strings
		div1Ind = curLine.indexOf(divStr);
		div2Ind = curLine.indexOf(divStr,div1Ind+1);
		endInd = curLine.indexOf(endStr);
		if ((div1Ind == -1) | (div2Ind == -1) | (endInd == -1)) { return false; }
		curLabel = curLine.substring(0,div1Ind).trim();
		curActiveStr = curLine.substring(div1Ind+1,div2Ind).trim();
		curValueStr = curLine.substring(div2Ind+1,endInd);
		// make value string a double
		try {
			Double valD = Double.valueOf(curValueStr);
			curValue = valD.doubleValue();
		} catch (java.lang.NumberFormatException nfe) {
			talkln("Invalid parameter value, " + curValueStr + " loaded from " + fromArg + "... exiting");
			System.exit(0);
		}
		// set boolean
		if (curActiveStr.equals("false")) { curActive = false; } else { curActive = true; }
		// find the matching parameter label from static list of parameters
		boolean matchFound = false;
		for (int i=0;i<Parameter_.paramCt;i++) {
			if (Parameter_.theParams[i].label.equals(curLabel)) {
				Parameter_.theParams[i].setValue(curValue);
				Parameter_.theParams[i].setActive(curActive);
				matchFound = true;
				if (Parameter_.theParams[i].type == Parameter_.BOOLEAN) {
					talkln ("     Parameter assigned: " + curLabel + " set to " + String.valueOf(Parameter_.theParams[i].isActive()));
				} else if (Parameter_.theParams[i].type == Parameter_.INT) {
					talkln ("     Parameter assigned: " + curLabel + " set to " + String.valueOf(Parameter_.theParams[i].getIntValue()));
				} else {
					talkln ("     Parameter assigned: " + curLabel + " set to " + String.valueOf(Parameter_.theParams[i].getValue()));
				}
			}
			if (matchFound) { break; }
		}
		if (!matchFound) { 
			talkln ("     No match found for parameter label " + curLabel + " ... you best check that out!"); 
		}
		return matchFound;
		
	}

	public static void loadParamConfig (File paramFile, boolean restartNeeded) {
		talkln ("Loading parameter values from " + paramFile.getAbsolutePath() + "...");
		int misunderstandings = 0;
		if ( ! paramFile.exists() || ! paramFile.canRead() ) {
			talkln ("Can't find or access " + paramFile.getAbsolutePath() + " ... check the file path and name");
		} else {
			try {
				FileReader fileR = new FileReader(paramFile);
				BufferedReader in = new BufferedReader(fileR);
				String curLine = in.readLine();
				String fromArg = paramFile.getAbsolutePath();
				while (curLine != null) {
					if (!curLine.startsWith(commentStr) & !curLine.startsWith(Info.appParamID) & (curLine.length() >= 1)) {
						boolean matchFound = loadParamLine(curLine,fromArg);
						if (!matchFound) { misunderstandings++; }
					}
					// next line
					curLine = in.readLine();
					
				}
			} catch (IOException ioe) { talkln ("An error trying to load param file"); }
		}
		talkln ("Done... there were " + String.valueOf(misunderstandings) + " misunderstandings.");
		if (restartNeeded) {
			System.out.print ("Restarting simulation... ");
			SimFil.restartRun(true);
			talkln("done.");
		}
		if (Env.remote) { 
			SimFil.setRunning(); 
		} else {
			ParamGui_.syncAllToParameters();
			SimFil_Graphics_.resetView();
		}
		
	}
	
	public static void loadParamsFromViewer (String viewStr) {
		talkln ("Loading parameter values from viewer...");
		String fromArg = "the viewer";
		int misunderstandings = 0;
		int bolIndex = 0;
		int eolIndex = 0;
		String curLine;
		int finalIndex = viewStr.lastIndexOf("\n");
		while (bolIndex < finalIndex) {
			eolIndex = viewStr.indexOf("\n",bolIndex);
	    	curLine = viewStr.substring(bolIndex,eolIndex);
	    	//System.out.println(curLine);
	    	if (!curLine.startsWith(commentStr) & !curLine.startsWith(Info.appParamID) & (curLine.length() >= 1)) {
				boolean matchFound = loadParamLine(curLine,fromArg);
				if (!matchFound) { misunderstandings++; }
			}
			// next line
			bolIndex = eolIndex+1;
		}
		talkln ("Done... there were " + String.valueOf(misunderstandings) + " misunderstandings.");
		System.out.print ("Restarting simulation... ");
		SimFil.restartRun(true);
		talkln("done.");
		if (Env.remote) { 
			SimFil.setRunning(); 
		} else {
			ParamGui_.syncAllToParameters();
			SimFil_Graphics_.resetView();
		}
		
	}
	
	public static void talkln (String info) {
		if (Env.remote & Env.logFiles) {
			//if (!outFileMade) { mkOutFile(); }
			//outPW.println(info);
			System.out.println(info);
		} else {
			System.out.println(info);
		}
	}
	
	public static void talk (String info) {
		if (Env.remote & Env.logFiles) {
			//if (!outFileMade) { mkOutFile(); }
			//outPW.print(info);
			System.out.print(info);
		} else {
			System.out.print(info);
		}
	}
}
