package simFil;
/*
 * SimFil
 */

/*  <SimFil - simulation of cytoskeletal filament flexibility by Pairwise Agent Interaction with Rational Superposition (PAIRS)>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.text.DecimalFormat;

import javax.swing.Timer;


public class SimFil {
	
	// graphics component
	static SimFil_Graphics_ flexFilGraphics;
	static TimeLoop timeLoop;
	
	// output formats
	static DecimalFormat deflectionFormat = new DecimalFormat ("0.0000");
	static DecimalFormat expFormat = new DecimalFormat ("0.000E0");
	
	public SimFil (String[] args) {
		parseArgs(args);
		
		if (Env.paramFile != null) { FileOps_.loadParamConfig(Env.paramFile, false); }
		if (Env.logFiles) { FileOps_.remoteParamConfigSave(); }
		
		if (!Env.remote) { 
			flexFilGraphics = new SimFil_Graphics_( );
			flexFilGraphics.initialize( );
			flexFilGraphics.buildUniverse( );
			flexFilGraphics.showFrame( ); 
		}
		
		// make Things
		Bug.makeInitialBug();
		makeInitialThings();
		
		// make the time loop instance
		timeLoop = new TimeLoop();
		timeLoop.start();
	}
	
	protected void parseArgs( String[] args ) {
		for ( int i = 0; i < args.length; i++ ) {
			if (args[i].equals("-help")) {
				talkln (" Command line options for ParMSpindle:");
				talkln (" -help: print this message");
				talkln (" -r  : run the code 'remotely', without any graphics output");
				talkln (" -pf -paramfile -paramFile (file) : load parameter values from specified file");
				talkln (" -o -out (directory) : create specified directory... all log, histogram, and qk files will be saved there");
				talkln (" -lf -logfile -logFile (directory) : create specified directory and save all output files to there");
				System.exit(0);
			}
			
			if ( args[i].equals( "-r" )) {
				Env.remote = true;
				Env.paused = false;
			}
			
			if ((args[i].equals("-paramfile")) | (args[i].equals("-pf")) | (args[i].equals("-paramFile"))) {
				File paramFile = new File(args[i+1]);
				if (paramFile.exists()) {
					Env.paramFile = paramFile;
				} else {
					talkln ("Specified param file can't be found... check path, etc");
					System.exit(0);
				}
			}
			
			if (args[i].equals("-ic")) {
				File icFile = new File(args[i+1]);
				if (icFile.exists()) {
					Env.fromQKStatePath = icFile.getAbsolutePath();
				} else {
					talkln ("Specified ic file can't be found... check path, etc");
					System.exit(0);
				}
			}
			
			if ((args[i].equals("-o")) | (args[i].equals("-out")) | (args[i].equals("-outfile"))) {
				File outFolder = new File(args[i+1]);
				File altOutFolder = new File (args[i+1]);
				int j=1;
				while (altOutFolder.isDirectory()) {
					System.out.println (altOutFolder.getName() + " exists.... keeping file name BUT changing directory name");
					altOutFolder = new File(args[i+1] + "." + String.valueOf(j));
					j++;
				}
				Env.outFileName = outFolder.getName();
				altOutFolder.mkdir();
				// directory for the log files
				File logSubFolder = new File(altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-LOG");
				Env.logFolderPath = logSubFolder.getAbsolutePath();
				logSubFolder.mkdir();
				Env.logFiles = true;
				// directory for the source files
				File srcSubFolder = new File (altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-SRC");
				Env.srcFilePath = srcSubFolder.getAbsolutePath();
				srcSubFolder.mkdir();
				// directory for the qk files
				File qkSubFolder = new File(altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-QK");
				Env.qkFilePath = qkSubFolder.getAbsolutePath();
				qkSubFolder.mkdir();
				Env.toQKFile = true;
				//*******
				
			}
			
			if (args[i].equals("-outMade")) {
				File outFolder = new File(args[i+1]);
				Env.outFileName = outFolder.getName();
	
				// directory for the log files
				File logSubFolder = new File(outFolder.getAbsolutePath() + File.separator + Env.outFileName + "-LOG");
				Env.logFolderPath = logSubFolder.getAbsolutePath();
				logSubFolder.mkdir();
				Env.logFiles = true;
				// directory for the qk files
				File qkSubFolder = new File(outFolder.getAbsolutePath() + File.separator + Env.outFileName + "-QK");
				Env.qkFilePath = qkSubFolder.getAbsolutePath();
				qkSubFolder.mkdir();
				Env.toQKFile = true;
				//*******
				
			}
			
			if ((args[i].equals("-logfile")) | (args[i].equals("-lf")) | (args[i].equals("-logFile"))) {
				File outFolder = new File(args[i+1]);
				File altOutFolder = new File (args[i+1]);
				int j=1;
				while (altOutFolder.isDirectory()) {
					System.out.println (altOutFolder.getName() + " exists.... keeping file name BUT changing directory name");
					altOutFolder = new File(args[i+1] + "." + String.valueOf(j));
					j++;
				}
				altOutFolder.mkdir();
				Env.outFileName = outFolder.getName();
				Env.logFolderPath = altOutFolder.getAbsolutePath();
				// make a directory for the source files
				File srcdir = new File (altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + ".SRC");
				//srcdir.mkdir();
				Env.srcFilePath = srcdir.getAbsolutePath();
				//*******
				Env.logFiles = true;
			}
			
		}
	}
	
	class TimeLoop extends Thread {
	
		public void run() {
			if (!Env.remote) {
				SimFil_Graphics_.updateBugScene();		// draw once at the beginning of the simulation
				startUpdateTimer();
			}
			doLoop();
			System.exit(0);
		}
	}
	
	public static void doLoop() {
		boolean paintedThisStep;
		int drawCounter = 0;
		int toFileCounter = 0;
		int remoteOutCounter = 0;
		int collisionCkCounter = 0;
		int ckElasticityCounter = 0;
		int ckPersistenceCounter = 0;
		
		while (Env.simulationTime <= Env.runTime.getValue()) {
			if (Env.paused) { 
				if (Env.viewRotation) {
					flexFilGraphics.rotateView();  // let's us set up a good looking rotation before starting movie
					try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
				} else {
					try { Thread.sleep(1000); } catch (InterruptedException e) { talkln ("error sleeping"); }
				}
			}
			if (!Env.paused) {
				synchronized(Env.safeO) {
					// Plasmid Collisions
					if (Env.simulationTime != 0) { PSphere.checkPlasmidCollision(); }
					 // Brownian Motion
					Thing.brownianMotionForAll();
					// all Things step forward in time
					 for (int i=0;i<Thing.thingCt;i++) {
						Thing.theThings[i].step();
					 }
					
					//update counters and flags
					paintedThisStep = false;
					Env.counter++;
					Env.simulationTime += Env.deltaT.getValue();
					remoteOutCounter++;
					collisionCkCounter++;
					ckElasticityCounter++;
					ckPersistenceCounter++;
					
					if (!Env.remote) { 
						//parMGraphics.setTopLabels();
						drawCounter++;
						toFileCounter++;
						
						if ((Env.paintOn) & (drawCounter >= Env.drawInterval.getIntValue() | Env.simulationTime == 0)) {
							if ((Env.viewRotation) & (!Env.toFile)) { flexFilGraphics.rotateView(); }
							flexFilGraphics.updateBugScene();
							paintedThisStep = true;
							drawCounter = 0;
						}
						
						if ((Env.toFile) & (toFileCounter >= Env.toFileInterval.getIntValue() | flexFilGraphics.theCanvas.firstRender())) {
							if (!paintedThisStep) {
								if (Env.viewRotation) { flexFilGraphics.rotateView(); }
								flexFilGraphics.updateBugScene();
								paintedThisStep = true;
								try { Thread.sleep(500); } catch (InterruptedException e) { talkln ("error sleeping"); }
							}
							flexFilGraphics.captureImage(flexFilGraphics.theCanvas);
							flexFilGraphics.waitOnCapture(flexFilGraphics.theCanvas);
							toFileCounter = 0;
							try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
						}
						
						/*if (Env.cantileverElasticityTst & (ckElasticityCounter >= Env.ckElasticityInterval)) {
							elasticityCk();
							ckElasticityCounter = 0;
						}*/
						
						if (Env.testingPersistenceLength & (ckPersistenceCounter >= Env.ckPersistenceInterval)) {
							FilSegment.persistenceTrack();
							ckPersistenceCounter = 0;
						}
						
					} else {
						
						/*if (Env.cantileverElasticityTst & (ckElasticityCounter >= Env.ckElasticityInterval)) {
							elasticityCk();
							ckElasticityCounter = 0;
						}*/
						
						if (Env.simpleSupportElasticityTst & Env.simulationTime > 100) {
							FilSegment.simpleSupportRelaxActions();
						}
						
						if (Env.testingPersistenceLength & (ckPersistenceCounter >= Env.ckPersistenceInterval)) {
							FilSegment.persistenceTrack();
							ckPersistenceCounter = 0;
						}
						
					}
					
					// remove Things AFTER the graphical update... can detach graphics objects of dead things this way
					Thing.removeDeadThings();		// get rid of dead things
					// stochastically create new filaments
					
				}
			}
		}
	}
	
	public static void makeInitialThings() {
		if (Env.simpleSupportElasticityTst) { simpleSupportElasticityTuning(); }
		if (Env.cantileverElasticityTst) { cantileverElasticityTuning(); }
		if (Env.persistenceTst) { persistenceLengthTst(); }
		if (Env.filUnderTraction) { filamentUnderTraction(); }
		if (Env.sphereBetweenPlanes) { sphereBetweenPlanes(); }
		if (Env.spheresUnderForceField) { spheresUnderForceField(); }
		if (Env.spheresSqueezeFit) { spheresSqueezeFit(); }
		if (Env.spheresLinked) { spheresLinked(); }
		
	}
	
	public static void restartRun (boolean newParamsLoaded) {
		setPaused();
		synchronized(Env.safeO) {
			FilSegment.removeAll();
			PSphere.removeAll();
			Thing.removeDeadThings();
			Thing.thingCt = 0;	// no Things left
			Env.simulationTime = 0;
			
			if (!Env.remote) { Thing.theBug.G.detach(); }
			Thing.theBug = null;
			Bug.makeInitialBug();
			
			if (Env.cantileverElasticityTst | Env.simpleSupportElasticityTst) { Env.resetElasticityTstValues(); }
			if (Env.testingPersistenceLength) { Env.resetPersistenceTstValues(); }
			
			makeInitialThings();
			
			if (!Env.remote) {
				if (newParamsLoaded) {
					if (!flexFilGraphics.canvasSizeCurrent()) { flexFilGraphics.makeCanvasSizeCurrent(); }
				} else {
					flexFilGraphics.updateCanvasSizeParams();
				}
				SimFil_Graphics_.updateBugScene();
				SimFil_Graphics_.updateAllControls();
				SimFil_Graphics_.resetTuningTestsState();
				SimFil_Graphics_.theCanvas.resetFileCt();
			}
			
		}
	}
	
	public static void setRunning () {
		Env.paused = false;
		if (!Env.remote) {
			flexFilGraphics.pausedMenuItem.setState(false);
			flexFilGraphics.runMenuItem.setState(true);
		}
	}
	
	public static void setPaused () {
		Env.paused = true;
		if (!Env.remote) {
			flexFilGraphics.pausedMenuItem.setState(true);
			flexFilGraphics.runMenuItem.setState(false);
		}
	}
	
	public static void startUpdateTimer() {
		int delay = 1000; //milliseconds
		ActionListener taskPerformer = new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				flexFilGraphics.updateAllControls();
		    }
		};
		new Timer(delay, taskPerformer).start();
	}
	
	public static void elasticityCk () {
		Env.curDeltaD = Env.lastDeflection - Env.currentDeflection;
		Env.curDistFromExpected = Env.expectedDeflection - Env.currentDeflection;
		
		if (Math.abs(Env.curDistFromExpected) > Math.abs(Env.lastDistFromExpected)) {	// overshoot
			FileOps_.talk ("Overshoot... ");
			changeElasticity();
		}
		if (Math.abs(Env.curDeltaD) < Env.ckElasticitySmallChange) { // if sufficiently static then check
			FileOps_.talk ("Static deflection achieved...");
			if (Math.abs(Env.curDistFromExpected) < Env.valOKChange) {	// close enough to expected deflection
				FileOps_.talkln (" done with this length");
				Env.elasReportString = "Final k value with filament length of " + deflectionFormat.format(Env.filamentLength.getValue()) + " is " + expFormat.format(Env.filTorqSpring.getValue());
				FileOps_.talkln (" **** " + Env.elasReportString);
				FileOps_.talkln ("");
				Env.curMultipleOfSeg ++;
				Env.simulationTime = 0;
				Env.resetElasticityTstValues();
				if (Env.filamentLength.getValue() > Env.tstFilLengthMax) { 
					FileOps_.talkln ("Done with elasticity test");
					System.exit(0); 
				}
			} else {
				changeElasticity();
			}
		}
		Env.lastDeflection = Env.currentDeflection;
		Env.lastDistFromExpected = Env.curDistFromExpected;
	}
	
	public static void changeElasticity () {
		if (Math.abs(Env.currentDeflection) > Math.abs(Env.expectedDeflection)) { 
			if (!Env.lastEChangePlus) {	// only halve elasticity change if we have switched direction of change
				Env.curDeltaE = Env.curDeltaE/2;
				Env.lastEChangePlus = true;
			}
			Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()+Env.curDeltaE); 
			FileOps_.talkln (" increasing spring to " + Env.filTorqSpring.getValue());
		} else {
			if (Env.lastEChangePlus) {
				Env.curDeltaE = Env.curDeltaE/2;
				Env.lastEChangePlus = false;
			}
			Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()-Env.curDeltaE);
			FileOps_.talkln (" decreasing spring to " +  + Env.filTorqSpring.getValue());
		}
	}
	
	public static void persistenceLengthTst() {
		Env.bugOn = false;
		Env.testingPersistenceLength = true;
		Env.resetPersistenceTstValues();
		Env.plasmidBrownianMotionOff = true;
		Env.filBrownianMotionOff = false;
		int monCt = Env.persistenceSegNumber*Env.stdSegLength.getIntValue();
		Pt3D loc = new Pt3D(0,0,0);
		Pt3D ang = new Pt3D(1,0,0);
		FilSegment lpFil = new FilSegment (loc,ang,-1,monCt,false);
	
		// spit out expected diffusivities for rigid rod
		double lpFilLengthM = 1.0e-6*lpFil.length; // in meters
		double radiusM = Env.filamentRadius.getValue()*1.0e-6;
		double denomLogTerm = Math.log(lpFilLengthM/(2*radiusM));	//dimensionless
		//double xDiff = (Env.Boltz*Env.tempK)/(2*Math.PI*Env.aeta.getValue()*lpFilLengthM)/(denomLogTerm + aParallel);
		//double yDiff = (Env.Boltz*Env.tempK)/(4*Math.PI*Env.aeta.getValue()*lpFilLengthM)/(denomLogTerm + aOrthog);
		//FileOps.talkln ("Expected diffusivities (if a rigid rod): " + xDiff + "," + yDiff + "," + yDiff);
		
	}
	
	public static void cantileverElasticityTuning() {
		Env.bugOn = false;
		Env.cantileverElasticityTst = true;
		Env.relaxWriteInterval = (int)3e3;
		Env.plasmidBrownianMotionOff = true;
		Env.filBrownianMotionOff = true;
		Env.resetElasticityTstValues();
		int monCt = (int)(Env.filamentLength.getValue()/(Env.monomerDiameter.getValue()/2)-1);
		Pt3D loc = new Pt3D(0,0,0);
		Pt3D ang = new Pt3D(1,0,0);
		FilSegment fil1 = new FilSegment (loc,ang,-1,monCt,false);
	
		Pt3D p1Loc = Pt3D.Add(fil1.end1,0.95*Env.sphRadius.getValue(),fil1.uVecR);
		PSphere p1 = new PSphere (p1Loc);
		p1.immobile = true;
	}
	
	public static void simpleSupportElasticityTuning() {
		Env.bugOn = false;
		Env.simpleSupportElasticityTst = true;
		Env.plasmidBrownianMotionOff = true;
		Env.filBrownianMotionOff = true;
		Env.resetElasticityTstValues();
		int monCt = (int)(Env.filamentLength.getValue()/(Env.monomerDiameter.getValue()/2)-1);
		Pt3D loc = new Pt3D(0,0,0);
		Pt3D ang = new Pt3D(1,0,0);
		FilSegment fil1 = new FilSegment (loc,ang,-1,monCt,false);
	
		Pt3D p1Loc = Pt3D.Add(fil1.end1,0.95*Env.sphRadius.getValue(),fil1.uVecR);
		PSphere p1 = new PSphere (p1Loc);
		p1.immobile = true;
		
		Pt3D p2Loc = Pt3D.Add(fil1.end2,0.95*Env.sphRadius.getValue(),fil1.uVec);
		PSphere p2 = new PSphere (p2Loc);
		p2.xSlider =true;
		
	}
	
	public static void filamentUnderTraction() {
		Env.bugOn = false;
		Env.filUnderTraction = true;
		Env.plasmidBrownianMotionOff = true;
		Env.filBrownianMotionOff = true;
		int monCt = (int)(Env.filamentLength.getValue()/(Env.monomerDiameter.getValue()/2)-1);
		Pt3D loc = new Pt3D(0,0,0);
		Pt3D ang = new Pt3D(1,0,0);
		new FilSegment (loc,ang,-1,monCt,false);
	}
	
	public static void sphereBetweenPlanes () {
		Env.bugOn = false;
		Env.plasmidBrownianMotionOff = true;
		Env.drawInterval.setValue(1);
		double plasWallRadius = Env.wallSphRadius.getValue();
		
		PSphere midPlas = new PSphere(new Pt3D(Env.midSphXOffset.getValue(),Env.midSphYOffset.getValue(),Env.midSphZOffset.getValue()));
		PSphere topPlas = new PSphere (new Pt3D(0,plasWallRadius+0.015,0));
		PSphere bottomPlas = new PSphere (new Pt3D(0,-plasWallRadius-0.015,0));
		
		midPlas.setCustomPlas(Env.midSphRadius.getValue(),-1,-1);
		topPlas.setCustomPlas(plasWallRadius, 1e-100, 1e-100);
		topPlas.setImmobile();
		bottomPlas.setCustomPlas(plasWallRadius, 1e-100, 1e-100);
		bottomPlas.setImmobile();
		
	}
	
	public static void spheresSqueezeFit () {
		Env.bugOn = true;
		Env.plasmidBrownianMotionOff = true;
		Env.drawInterval.setValue(1);
		int plasNum = Env.numSqueezed.getIntValue();
		PSphere curPlas;
		for (int i=0;i<plasNum;i++) {
			curPlas = new PSphere(Thing.theBug.rdmPtInsideBug());
			curPlas.setCustomPlas(Env.radSqueezed.getValue(),-1,-1);
		}
	}
	
	public static void spheresUnderForceField () {
		Env.bugOn = true;
		Env.plasmidBrownianMotionOff = true;
		Env.drawInterval.setValue(1);
		int plasNum = Env.numInBucket.getIntValue();
		PSphere curPlas;
		for (int i=0;i<plasNum;i++) {
			curPlas = new PSphere(Thing.theBug.rdmPtInsideBug());
			curPlas.setCustomPlas(Env.radInBucket.getValue(),1e-12,1e1);
		}
	}
	
	public static void spheresLinked () {
		Env.bugOn = false;
		Env.plasmidBrownianMotionOff = true;
		Env.drawInterval.setValue(1);
		int plasNum = Env.numAttracted.getIntValue();
		PSphere curPlas;
		for (int i=0;i<plasNum;i++) {
			curPlas = new PSphere(Thing.theBug.rdmPtInsideBug());
			curPlas.setCustomPlas(Env.radAttracted.getValue(),-1,-1);
		}
	}
	
	private static void talkln (String info) {
		System.out.println(info);
	}
	
	private static void talk (String info) {
		System.out.print(info);
	}
}
